//	WinDjView
//	Copyright (C) 2004-2007 Andrew Zhezherun
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License version 2
//	as published by the Free Software Foundation.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	http://www.gnu.org/copyleft/gpl.html

// $Id: NavPane.h,v 1.15 2007/04/11 15:46:33 zhezherun Exp $

#pragma once

#include "Drawing.h"


// CNavPaneWnd

class CNavPaneWnd : public CWnd, public Observer
{
	DECLARE_DYNCREATE(CNavPaneWnd)

public:
	CNavPaneWnd();
	virtual ~CNavPaneWnd();

// Attributes
public:
	enum
	{
		s_nTabsWidth = 20,
		s_nTabSize = 20,
		s_nLeftMargin = 5,
		s_nTopMargin = 22,
		s_nMinExpandedWidth = 100
	};

// Operations
public:
	int AddTab(const CString& strName, CWnd* pWnd);
	int GetTabIndex(CWnd* pTabContent);
	void ActivateTab(CWnd* pTabContent, bool bExpand = true);
	void ActivateTab(int nTab, bool bExpand = true);
	void SetTabName(CWnd* pTabContent, const CString& strName);
	void SetTabName(int nTab, const CString& strName);
	void SetTabBorder(CWnd* pTabContent, bool bDrawBorder);
	void SetTabBorder(int nTab, bool bDrawBorder);

	virtual void OnUpdate(const Observable* source, const Message* message);

// Implementation
protected:
	struct Tab
	{
		CString strName;
		CRect rcTab;
		CWnd* pWnd;
		bool bHasBorder;
	};
	vector<Tab> m_tabs;
	CFont m_font;
	CFont m_fontActive;
	int m_nActiveTab;
	void DrawTab(CDC* pDC, int nTab, bool bActive);
	void UpdateCloseButton(bool bLButtonDown);
	void UpdateCloseButtonImpl(bool bPressed, bool bActive, const CRect& rcButton);
	int GetTabFromPoint(CPoint point);
	void UpdateTabContents();
	bool PtInTab(int nTab, CPoint point);
	CToolTipCtrl m_toolTip;
	CImageList m_imgClose;

	COffscreenDC m_offscreenDC;

	bool m_bCloseActive;
	bool m_bClosePressed;
	bool m_bDragging;

	// Generated message map functions
	afx_msg void OnPaint();
	afx_msg void OnWindowPosChanged(WINDOWPOS FAR* lpwndpos);
	afx_msg BOOL OnEraseBkgnd(CDC* pDC);
	afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnLButtonUp(UINT nFlags, CPoint point);
	afx_msg void OnMouseMove(UINT nFlags, CPoint point);
	afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	afx_msg void OnDestroy();
	afx_msg void OnSetFocus(CWnd* pOldWnd);
	afx_msg int OnMouseActivate(CWnd* pDesktopWnd, UINT nHitTest, UINT message);
	virtual BOOL PreTranslateMessage(MSG* pMsg);
	virtual void PostNcDestroy();
	DECLARE_MESSAGE_MAP()
};
